
/****************************************************************************
** pc38nice.c
****************************************************************************
*/

#define EXTERN extern

/*------------------------
** Include files
**------------------------
*/
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <sys/ioctl.h>

#include <pmac.h>

/*----------------------------------------------------------------------
**  pmac_sleep() - Sleeps in driver for ms.
**----------------------------------------------------------------------
*/
int pmac_msleep( int fd, int ms )
{
   struct pmac_parms_t p;
   p.l[0] = ms;
   ioctl( fd, PMAC_IOCTL_MSLEEP, &p );
   return p.ret;
}

/*----------------------------------------------------------------------
**  pmac_string_reverse() - Test user_copy from/to by changing a user sting
**----------------------------------------------------------------------
*/
int pmac_string_reverse( int fd, char * buf )
{
   struct pmac_parms_t p;
   p.p[0] = buf;              // pointer to string
   p.l[0] = strlen(buf);      // chars to reverse
   ioctl( fd, PMAC_IOCTL_STRING_REVERSE, &p );
   return p.ret;
}

/*----------------------------------------------------------------------
**  pmac_reg_cmd() - command IO via the pci register interface
**----------------------------------------------------------------------
*/
int pmac_reg_cmd( int fd, char * cmd, char * rpy, int rpy_size )
{
   struct pmac_parms_t p;
   p.p[0] = cmd;              // command 
   p.l[0] = strlen(cmd);      // num char in command.
   p.p[1] = rpy;              // reply
   p.l[1] = rpy_size;         // sizeof(reply)
   ioctl( fd, PMAC_IOCTL_REG_CMD, &p );
   return p.ret;
}

/*----------------------------------------------------------------------
**  pmac_mem_puts() - write a command to the DPR interface
**----------------------------------------------------------------------
*/
int pmac_mem_puts( int fd, char * cmd )
{
   struct pmac_parms_t p;

   p.p[0] = cmd;              // command 
   p.l[0] = strlen(cmd);      // num char in command.
   ioctl( fd, PMAC_IOCTL_MEM_PUTS, &p );
   return p.ret;
}

/*----------------------------------------------------------------------
**  pmac_mem_gets() - reads a command from the DPR interface
**----------------------------------------------------------------------
*/
int pmac_mem_gets( int fd, char * rpy, int rpy_size )
{
   struct pmac_parms_t p;

   p.p[0] = rpy;              // reply
   p.l[0] = rpy_size;         // sizeof(reply)
   ioctl( fd, PMAC_IOCTL_MEM_GETS, &p );
   return p.ret;
}

/*----------------------------------------------------------------------
**  pmac_mem_flush() - flush DPR interface buffers
**----------------------------------------------------------------------
*/
int pmac_mem_flush( int fd )
{
   struct pmac_parms_t p;

   ioctl( fd, PMAC_IOCTL_MEM_FLUSH, &p );
   return p.ret;
}

/*----------------------------------------------------------------------
**  pmac_mem_cmd() - mem based IO 
**----------------------------------------------------------------------
*/
int pmac_mem_cmd( int fd, char * cmd, char * rpy, int rpy_size )
{
   struct pmac_parms_t p;
   p.p[0] = cmd;              // command 
   p.l[0] = strlen(cmd);      // num char in command.
   p.p[1] = rpy;              // reply
   p.l[1] = rpy_size;         // sizeof(reply)
   ioctl( fd, PMAC_IOCTL_MEM_CMD, &p );
   return p.ret;
}

/*----------------------------------------------------------------------
**  pmac_mem_getblk( ) - memcpy from PMAC io memory to host.
**----------------------------------------------------------------------
*/
int pmac_mem_getblk( int fd, void * dest, int src, int nbytes )
{
   struct pmac_parms_t p;
   p.p[0] = dest;            // memcpy destination
   p.l[0] = src;             // io offset of source
   p.l[1] = nbytes;          // number of bytes to copy
   ioctl( fd, PMAC_IOCTL_MEM_GETBLK, &p );
   return p.ret;
}

/*----------------------------------------------------------------------
**  pmac_mem_putblk( ) - memcpy from host to PMAC io memory.
**----------------------------------------------------------------------
*/
int pmac_mem_putblk( int fd, int dest, void * src, int nbytes )
{
   struct pmac_parms_t p;
   p.l[0] = dest;             // io offset addess to write to
   p.p[0] = src;              // host data (source)
   p.l[1] = nbytes;           // number of bytes to copy
   ioctl( fd, PMAC_IOCTL_MEM_PUTBLK, &p );
   return p.ret;
}

/*----------------------------------------------------------------------
**  pmac_mem_get_motor_data() - Gets motor data from DPRAM
**----------------------------------------------------------------------
*/
int pmac_mem_get_motor_data( int fd, void * dest, long * timestamp, int m0, int nmotors)
{
   struct pmac_parms_t p;
   p.p[0] = dest;     // address of user buffer
   p.l[0] = m0;       // offset for motor 0 - use PMAC_DPR_MOTORx
   p.l[1] = nmotors;  // number of motor block to copy
   p.l[2] = 0;        // driver returns timestamp here
   ioctl( fd, PMAC_IOCTL_MEM_GET_MOTOR_DATA, &p );
   * timestamp = p.l[2];
   return p.ret;
}

/*----------------------------------------------------------------------
**  pmac_t1() - test1
**----------------------------------------------------------------------
*/
int pmac_t1( int fd )
{
   struct pmac_parms_t p;
   ioctl( fd, PMAC_IOCTL_T1, &p );
   return p.ret;
}

/*----------------------------------------------------------------------
**  pmac_t2() - test2
**----------------------------------------------------------------------
*/
int pmac_t2( int fd )
{
   struct pmac_parms_t p;
   ioctl( fd, PMAC_IOCTL_T2, &p );
   return p.ret;
}

/*----------------------------------------------------------------------
**  dpr2double() - converts DPR float (8 bytes) to host double
**     return mastissa/2^35 * 2^(exp-2047)
**----------------------------------------------------------------------
*/
double dprFloat2Double( long first_word, long second_word )
{
   double m;
   int exp;

   m = (double) second_word * 4096.0 + (double) ((first_word>>12) & 0x00fffL);
   if( m == 0.0 ) return 0.0;

   exp = (first_word & 0x0fffL) - 2082;   // 2082 = 20470+35
   return  m * pow(2.0, (double)exp);
}

/*----------------------------------------------------------------------
**  dprLongFixed() - converts DPR long word (8 bytes) to host double
**----------------------------------------------------------------------
*/
double dprLongFixed( long d[], double scale )
{
   /*
   **  floating point caculation of [(d[1] << 24) + (d[0] 0xffffff)] * scale
   */
   return (((double)d[1] * 16777216.0) + (double)(d[0] & 0x00ffffff)) * scale;
}


/*----------------------------------------------------------------------
**  dprShortFixed() - converts DPR short word (4 bytes) to long 
**----------------------------------------------------------------------
*/
long dprShortFixed( long word )
{
   if( 0x00800000 & word )
      return (word | 0xff000000);  // extend size bits & return 
   else
      return (word & 0x00ffffff);  // return 24 bits from word
}

/*----------------------------------------------------------------------
**  dprShortU() - converts DPR short word (4 bytes) to unsigned long 
**----------------------------------------------------------------------
*/
ulong dprShortU( ulong word )
{
   return (word & 0x00ffffff);  // return 24 bits from word
}

