
#define EXTERN
#define MAIN     1
 
/*--------------------------
**  include files
**--------------------------
*/
#include <sys/types.h>
#include <sys/time.h>
#include <ctype.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include <rpc/rpc.h>

#include <semaphore.h>
#include <mqueue.h>

#include <pmac.h>
#include <pm.h>

/*-----------------------------
** Non-standard include files
**-----------------------------
*/

#include <slamac.h>

#include "t3lib.h"

#define DAS2CNT 102.4      // for Lab System Computer

/*---------------------------
** function prototypes
**---------------------------
*/
int cmd_execute( char * cmd_buf);

int do_ape_init( char *par );
int do_exe( char *par );
int do_iom( char *par );
int do_ior( char *par );
int do_m1( char *par );
int do_mflush( char *par );
int do_mget( char *par );
int do_mput( char *par );
int do_msleep( char *par );
int do_show_md( char *par );
int do_show_plc0( char *par );
int do_show_vb( char *par );
int do_string_reverse( char *par );
int do_t1( char *par );
int do_t2( char *par );

int do_pm_init( char *par );
int do_pm_show( char *par );
int pm_show_print_motor_data( struct pm_minfo_t *m , char * name, int x, int y );

int do_help( char *par );
int do_quit( char *par );
int do_test( char *par );

ulong get_ulong( char *prompt, ulong cur_val, ulong min, ulong max );
long  get_long ( char *prompt, long cur_val, long min, long max );
double get_dbl( char *prompt, double cur_val, double min, double max );
void tm_start( void );
double tm_end( void );
void show_mem( void * mem, int nbytes, char * desc );

int sc_clrscr( int p );
int sc_clrln( int p );
int sc_gotoxy( int x, int y);
int sc_home(void );

int stio_cmd ( char * cmd, char * reply, int reply_size );

/*-------------------------------------
*  Global variables
*-------------------------------------
*/

int  Quit;
int  Fd;

struct PARSE_TABLE
{
   char * kw;          /* The opcode string  */
   int (* pfi)();      /* pfi is pointer to function returning an integer */
   char * desc;       /* command description */
};

struct PARSE_TABLE parse_table[] =
{
   /* BASIC DRIVER COMMANDS */
   { "ape.init",         do_ape_init,        "ape.init - sync PMAC pos to APE" },
   { "exe",              do_exe,             "exe Filename - send execute file to pmac" },
   { "iom",              do_iom,             "iom - command IO (DPR interface)" },
   { "ior",              do_ior,             "ior - command IO (register interface)" },
   { "m1",               do_m1,              "m1 - " },
   { "mflush",           do_mflush,          "mflush - mem_flush" },
   { "mput",             do_mput,            "mput - write a command to DPR" },
   { "mget",             do_mget,            "mget - read a reply from DPR" },
   { "msleep",           do_msleep,          "msleep - Test driver sleep function" },
   { "show.md",          do_show_md,         "show.md - Show Motor Data Reporting Buffer" },
   { "show.plc0",        do_show_plc0,       "show.plc0 - Show plc0 variables" },
   { "show.vb",          do_show_vb,         "show.vb - display some VB data " },
   { "string.reverse",   do_string_reverse,  "string.reverse - reverses a string" },
   { "t1",               do_t1,              "t1 - " },
   { "t2",               do_t2,              "t2 - " },

   /* T3 APPLICATION specific COMMANDS */
   { "pm.init",          do_pm_init,         "pm.init - Init PMAC for the T3 application " },
   { "pm.show",          do_pm_show,         "pm.show - display MotorInformation" },

   /* MISC COMMANDS */
   { "help",             do_help,            "Help - List commands" },
   { "Quit",             do_quit,            "Quit - quits app" },
   { "Test",             do_test,            "Test - test code" },
};

int num_parse_item = { sizeof(parse_table) / sizeof(struct PARSE_TABLE) };


/*---------------------------------
** main()
**---------------------------------
*/
int main( int argc, char * argv[])
{
   int i, err;
   extern char *optarg;
   extern int optind;
   char buf[80];

   /* parse options */
   err = 0;
   while((i = getopt(argc, argv, "")) != EOF)
   {
      switch(i)
      {
         case '?':
         err++;
         break;
      }
   }

   if( err )
   {
      fprintf(stderr, "Usage: %s [] \n", argv[0]);
      exit(0);
   }

   /*--------------------------------------
   **  open device driver 
   */
   if( (Fd = open("/dev/pmac", O_RDWR, 0x666)) < 0)
   {
      printf("Error opening pmac device driver \n");
      exit(0);
   }
   printf("opened /dev/pmac fd=%d \n", Fd );


   /* main loop for commands */
   Quit = FALSE;
   while( !Quit )
   {
      /* Get command */
      printf("Enter Command: ");
      gets( buf );

      cmd_execute( buf );
   }

   close( Fd );
   return 0;
}

/*-------------------------------------------------------------
** cmd_execute() - Executes a command line instruction
**-------------------------------------------------------------
*/
int cmd_execute( cmd_buf )
char * cmd_buf;
{
   char copy_buf[80];
   int error,
      found,
      i;

   char * kw,
        * par;
   /*
   **  Search table for keyword
   */
   strxcpy( copy_buf, cmd_buf, sizeof(copy_buf));

   kw  = strtok(copy_buf, " ");
   if( (kw == NULL) || (*kw=='#') )
      return( T3_ERR_NONE );

   if( (par = strtok( (char*)NULL, "")) != NULL )
      unpad( par, ' ' );

   /*
   **  Search table for keyword. If found call function.
   */
   error = T3_ERR_INV_KW;
   for(i=0, found = FALSE; i<num_parse_item && found == FALSE; )
   {
      if( !stricmp( kw, parse_table[i].kw ) )
      {
         found = TRUE;
         error = (parse_table[i].pfi)(par);
      }
      else
         i++;
   }
   /*
   **  Provide user feedback
   */

   if( error != T3_ERR_NONE )
   {
      printf("%s - %s\n", cmd_buf, t3_error_string(error) );
   }

   return( error );
}

/********************************************************************
**                    BASIC DRIVER COMMANDS                        **
********************************************************************/

/*----------------------------------------------------------------------
**  do_ape_init() - get the ape information and sync pmac apos
**----------------------------------------------------------------------
*/
int do_ape_init( char * par )
{
   int rc;
   char buf[80];
   double rad[2];
   long   step[2];
   char * str_ptr;

   /* get ape info from simtac */
   stio_cmd( "ape.get", buf, sizeof(buf));

   if( (rc = parseDouble_r( &rad[0], buf, " ", &str_ptr  ) ) != T3_ERR_NONE )
      return T3_ERR_INV_FORMAT;
   if( (rc = parseDouble_r( &rad[1], NULL, " ", &str_ptr  ) ) != T3_ERR_NONE )
      return T3_ERR_INV_FORMAT;

   step[0] = rad[0]*DR2AS*DAS2CNT;
   step[1] = rad[1]*DR2AS*DAS2CNT;

   printf("radians: %f %f \n", rad[0], rad[1] );
   printf("  steps: %ld %ld \n", step[0], step[1]);

   /* set pmac position to ape */
   pm_apos_set( Fd, step[0], step[1] ); 
   
   return T3_ERR_NONE;
}

/*----------------------------------------------------------------------
**  do_exe() - exe feeds text command to the MEM command buffer
**     Syntax: exe [echo]
**----------------------------------------------------------------------
*/
int do_exe( char * par )
{
   int rc;
   char cmd[256];
   char rpy[256];
   char * cptr;

   FILE * fp;

   if( NULL== ( fp = fopen( par, "rt") ) )
      return T3_ERR_FILE_READ; 

   while( NULL != fgets( cmd, sizeof(cmd), fp ) )
   {
      str_rws(cmd, sizeof(cmd));
      unpad(cmd, ' ');

      if( cmd[0] == ';' )       // comment line...continue
         continue; 
      if( strlen(cmd) < 1)       // blank line...continue
         continue; 

      /* remove any comments embedded in instruction line */
      if( (cptr = strchr( cmd, ';')) != NULL )
      {
         *cptr = 0;
         unpad( cmd, ' ' );
      }

      if( (rc = pmac_reg_cmd( Fd, cmd, rpy, sizeof(rpy) )) != T3_ERR_NONE)
      {
         printf("ERROR: pmac_reg_cmd()=%d. Aborting exe...\n", rc);
         break;
      }

      /* echo input/output only if their is a pmac reply */
      if( strlen(rpy) > 0 )
      {
         printf(">%s\n", cmd);
         printf("%s\n", rpy);
      }
   }

   fclose( fp );
   return T3_ERR_NONE;
}

/*----------------------------------------------------------------------
**  do_iom() - calls pmac:ioctl(PMAC_CMD_MEM_IO) - IO using mem interface
**----------------------------------------------------------------------
*/
int do_iom( char * par )
{
   int  rc, l, i;
   double esec;
   char buf[255];

   while( 1 )
   {
      printf("IOM> " ); fflush(stdout);
      fgets( buf, sizeof(buf), stdin );

      if( strlen(buf) <= 1)
          return T3_ERR_NONE;

      tm_start( );
      if( (rc = pmac_mem_cmd( Fd, buf, buf, sizeof(buf) )) != T3_ERR_NONE)
         printf("ERROR: pmac_mem_cmd()=%d\n", rc);
      esec = tm_end( );
      l = strlen(buf);

      // display return values
      for( i=0; i<l; i++ )
         printf("%c", (isspace(buf[i])?' ':buf[i]));
      printf("\n");

      // display as hex 
      for( i=0; i<l; i++ )
         printf("0x%x ", buf[i]);
      printf("esec=%.6f \n", esec);

   }
   return T3_ERR_NONE;
}

/*----------------------------------------------------------------------
**  do_ior() - calls pmac:ioctl(PMAC_CMD_REG_IO) - IO using register interface
**----------------------------------------------------------------------
*/
int do_ior( char * par )
{
   int  rc, l, i;
   double esec;
   char buf[255];

   while( 1 )
   {
      printf("IOR > " ); fflush(stdout);
      fgets( buf, sizeof(buf), stdin );

       if( strlen(buf) <= 1)
          return T3_ERR_NONE;

      tm_start( );
      if( (rc = pmac_reg_cmd( Fd, buf, buf, sizeof(buf) )) != T3_ERR_NONE)
         printf("ERROR: pmac_reg_cmd()=%d\n", rc);
      esec = tm_end( );
      l = strlen(buf);

      // display return values
      for( i=0; i<l; i++ )
         printf("%c", (isspace(buf[i])?' ':buf[i]));
      printf("\n");

      // display as hex 
      for( i=0; i<l; i++ )
         printf("0x%x ", buf[i]);
      printf("esec=%6.f \n", esec);
   }
   return T3_ERR_NONE;
}

/*----------------------------------------------------------------------
**  do_m1() - prototype for pmac_vb_start().
**----------------------------------------------------------------------
*/
int do_m1( char * par )
{
   int rc;

   rc = pm_vb_init( Fd );
   printf(" pm_vb_init() returns %d\n", rc );
   return 0;
}

/*----------------------------------------------------------------------
**  do_mflush() - calls pmac:ioctl(PMAC_IOCTL_MEM_FLUSH)  
**     Flush DPR interface
**----------------------------------------------------------------------
*/
int do_mflush( char * par )
{
   int rc;

   rc = pmac_mem_flush( Fd );
   printf(" pmac_mem_flush() returns %d \n", rc);
   return 0;
}

/*----------------------------------------------------------------------
**  do_mput() - calls pmac:ioctl(PMAC_IOCTL_MEM_PUTS)  
**     write a command using DPR interface
**----------------------------------------------------------------------
*/
int do_mput( char * par )
{
   if (par==NULL )
      return T3_ERR_INV_RNG;

   pmac_mem_puts( Fd, par );
   return T3_ERR_NONE;
}

/*----------------------------------------------------------------------
**  do_mget() - calls pmac:ioctl(PMAC_IOCTL_MEM_GETS)  
**     Reads a reply from DPR interface
**----------------------------------------------------------------------
*/
int do_mget( char * par  )
{
   int rc, l, i;
   double esec;
   char reply[256];

   tm_start( );
   rc = pmac_mem_gets( Fd, reply, sizeof(reply) );
   esec = tm_end( );

   printf("Reply %d %.6f [%s]\n", rc, esec, reply);
   if( rc < 0 )
   {
      printf(" pmac_mem_gets() error %d [%s]\n", rc, reply);
      if( rc == -ETIMEDOUT ) printf("ETIMEOUT\n");
      if( rc == -ENODATA ) printf("ENODATA\n");
      if( rc == -EAGAIN ) printf("EAGAIN\n");
   }

   l = strlen(reply);
   printf("  %d:", l);
   for( i=0; i<l; i++ )
      printf("0x%x ", reply[i]);
   printf("\n");

   return T3_ERR_NONE;
}

/*----------------------------------------------------------------------
**  do_msleep() - calls pmac:ioctl(PMAC_CMD_MSLEEP) - sleeps in driver
**----------------------------------------------------------------------
*/
int do_msleep( char * par )
{
   int rc, i;
   long ms,  
        loops;
   struct timespec  start, end;

   ms    = 1000;
   loops = 10;

   ms = get_long( "Time(ms) ", ms, 1, 10000);
   loops = get_long( "Loops ", loops, 1, 1000);

   for( i=0; i<loops; i++ )
   {
      clock_gettime( CLOCK_REALTIME, &start );
      rc = pmac_msleep( Fd, ms );
      clock_gettime( CLOCK_REALTIME, &end );
      printf("   %02d %3d %.03f \n", i, rc, t3_clock_elapse_sec(&start, &end));
      usleep(1000);
   }
   return 0;
}

/*----------------------------------------------------------------------
**  do_show_md() - Show ONLY Motor Data Reporting Buffer data.
**----------------------------------------------------------------------
*/
int do_show_md( char * par )
{
   int rc, loops, i;
   long timestamp, l;
   struct pmac_MRB_raw_t md_raw;
   struct pm_minfo_t mi;
   char bits[40];
   char * str_ptr;

   /* get star index */
   rc = parseInt_r( &loops, par, " ", &str_ptr  );
   if( rc != T3_ERR_NONE)
      loops = 1;

   /* initialize motor info with scaling data */
   pm_mi_init_defaults( &mi );

   for( i=0; i<loops; i++ )
   {
      sc_home();
      if( (i%10) == 0 )
         sc_clrscr(2);

      //---------------------------------------
      // get data from motor reporting buffer
      //---------------------------------------

      /* get raw data from Motor Reporting DPR */
      rc = pmac_mem_get_motor_data( Fd, &md_raw, &timestamp, PMAC_DPR_MOTOR0, 1);

      printf("loop %04d/%04d Timestamp: %ld \n", i, loops, timestamp); 
      show_mem( &md_raw, PMAC_DPR_MOTOR_NBYTES, "MtrRptBuf");

      /* copy/convert data from MotorReportingBuffer to motorInfo */ 

      mi.sstatus   = md_raw.sstatus;
      mi.mstatus   = md_raw.mstatus;

      mi.pdac      = dprShortFixed( md_raw.pdac ) >> 8;   // 24 bits to 16 bits. 
      mi.mpos      = dprLongFixed( md_raw.hwpos, 1.0/(mi.Ixx08*32.0));
      mi.apos      = dprLongFixed( md_raw.apos, 1.0/(mi.Ixx08*32.0) );
      mi.bpos      = dprLongFixed( md_raw.bpos, 1.0/(mi.Ixx08*32.0)  );
      mi.pferr     = dprLongFixed( md_raw.pepos, 1.0/(mi.Ixx08*32.0)  );
      mi.fvel      = dprShortFixed( md_raw.fvel ) / (mi.Ixx09*32.0) * (8388608.0/mi.I10) * 1000.0;

      //---------------------------------------
      // get data from read variable buffer 
      //---------------------------------------

      //---------------------------------------
      // display motor info
      //---------------------------------------

      memcpy( &l, &mi.sstatus, sizeof(l));
      printf("   sstatus 0x%08lx : %s\n", l, btosl( bits, l));
      memcpy( &l, &mi.mstatus, sizeof(l));
      printf("   mstatus 0x%08lx : %s\n", l, btosl( bits, l));

      printf("      pdac %ld 0x%lx\n", mi.pdac, mi.pdac);
      printf("      mpos %.3f \n",  mi.mpos);
      printf("      apos %.3f \n",  mi.apos);
      printf("      bpos %.3f \n",  mi.bpos);
      printf("     pferr %.3f \n",  mi.pferr);
      printf("      fvel %.3f \n",  mi.fvel );
  
      sc_gotoxy( 1, 20 ); printf("_Servo_Status_BITS_");
      sc_gotoxy( 1, 21 ); printf("  rapid_spd_sel %d", mi.sstatus.rapid_spd_sel);
      sc_gotoxy( 1, 22 ); printf("   dac_sign_mag %d", mi.sstatus.dac_sign_mag);
      sc_gotoxy( 1, 23 ); printf("     sw_capture %d", mi.sstatus.sw_capture);
      sc_gotoxy( 1, 24 ); printf("     fe_capture %d", mi.sstatus.fe_capture);
      sc_gotoxy( 1, 25 ); printf("  handwheel_ena %d", mi.sstatus.handwheel_ena);
      sc_gotoxy( 1, 26 ); printf("        hw_mode %d", mi.sstatus.hw_mode);
      sc_gotoxy( 1, 27 ); printf("   phased_motor %d", mi.sstatus.phased_motor);
      sc_gotoxy( 1, 28 ); printf("     yenc_phase %d", mi.sstatus.yenc_phase);
      sc_gotoxy( 1, 29 ); printf("     user_servo %d", mi.sstatus.user_servo);
      sc_gotoxy( 1, 30 ); printf("     user_phase %d", mi.sstatus.user_phase);
      sc_gotoxy( 1, 31 ); printf("    home_search %d", mi.sstatus.home_search);
      sc_gotoxy( 1, 32 ); printf("  block_request %d", mi.sstatus.block_request);
      sc_gotoxy( 1, 33 ); printf("     limit_stop %d", mi.sstatus.limit_stop);
      sc_gotoxy( 1, 34 ); printf("  desired_vel_0 %d", mi.sstatus.desired_vel_0);
      sc_gotoxy( 1, 35 ); printf(" data_block_err %d", mi.sstatus.data_block_err);
      sc_gotoxy( 1, 36 ); printf("       dwelling %d", mi.sstatus.dwelling);
      sc_gotoxy( 1, 37 ); printf(" integrator_ena %d", mi.sstatus.integrator_ena);
      sc_gotoxy( 1, 38 ); printf("    run_program %d", mi.sstatus.run_program);
      sc_gotoxy( 1, 39 ); printf("      open_loop %d", mi.sstatus.open_loop);
      sc_gotoxy( 1, 40 ); printf("    amp_enabled %d", mi.sstatus.amp_enabled);
      sc_gotoxy( 1, 41 ); printf("       algo_ena %d", mi.sstatus.algo_ena);
      sc_gotoxy( 1, 42 ); printf("      pos_limit %d", mi.sstatus.pos_limit);
      sc_gotoxy( 1, 43 ); printf("      neg_limit %d", mi.sstatus.neg_limit);
      sc_gotoxy( 1, 44 ); printf("      activated %d", mi.sstatus.activated);

      sc_gotoxy(25, 20 ); printf("_Motor_Status_BITS_");
      sc_gotoxy(25, 21 ); printf("     in_position %d", mi.mstatus.in_position);
      sc_gotoxy(25, 22 ); printf("       warn_ferr %d", mi.mstatus.warn_ferr);
      sc_gotoxy(25, 23 ); printf("      fatal_ferr %d", mi.mstatus.fatal_ferr);
      sc_gotoxy(25, 24 ); printf("       amp_fault %d", mi.mstatus.amp_fault);
      sc_gotoxy(25, 25 ); printf("    backlash_dir %d", mi.mstatus.backlash_dir);
      sc_gotoxy(25, 26 ); printf("     amp_i2t_err %d", mi.mstatus.amp_i2t_err);
      sc_gotoxy(25, 27 ); printf("   integral_ferr %d", mi.mstatus.integral_ferr);
      sc_gotoxy(25, 28 ); printf(" triger_home_flg %d", mi.mstatus.triger_home_flg);
      sc_gotoxy(25, 29 ); printf("  phase_find_err %d", mi.mstatus.phase_find_err);
      sc_gotoxy(25, 30 ); printf("   home_complete %d", mi.mstatus.home_complete);
      sc_gotoxy(25, 31 ); printf("stopped_on_limit %d", mi.mstatus.stopped_on_limit);
      sc_gotoxy(25, 32 ); printf("     cs_assigned %d", mi.mstatus.cs_assigned);
      sc_gotoxy(25, 33 ); printf("          cs_def %d", mi.mstatus.cs_def);
      sc_gotoxy(25, 34 ); printf("       coord_sys %d", mi.mstatus.coord_sys);

      usleep(200000);
   }
   sc_gotoxy( 1, 50 );

   return T3_ERR_NONE;
}

/*----------------------------------------------------------------------
**  do_show_plc0() -  shows ONLY the open buffer PLC0 variables. 
**----------------------------------------------------------------------
*/
int do_show_plc0( char * par )
{
   int rc, i, loops;
   int addr;
   char * str_ptr;
   struct pm_VB_plc0_raw_t raw;
   struct pm_minfo_t  m1;
   struct pm_minfo_t  m2;

   /* get star index */
   rc = parseInt_r( &loops, par, " ", &str_ptr  );
   if( rc != T3_ERR_NONE)
      loops = 1;

   for( i=0; i<loops; i++ )
   {
      sc_home();
      if( (i%10) == 0 )
         sc_clrscr(2);
      printf("loop %04d/%04d \n", i, loops ); 

      /* get ob read data */
      addr = PMAC_VB_DATA_ADDR + (PM_VB_OPEN_BUF_OFFSET*4);
      rc = pmac_mem_getblk( Fd, &raw, addr, sizeof(raw));
      show_mem( &raw, sizeof(raw), "PLC0 Data ");

      printf(" addr 0x%x\n", addr );

      /* convert vb read raw data to minfo_t format */
      m1.timestamp_plc0 = dprShortU( raw.ts );       
      m1.period_plc0    = dprShortFixed( raw.period ); 
      m1.base   = dprShortFixed(raw.m1_base);
      m1.dac0   = dprShortFixed(raw.m1_dac0);
      m1.dac1   = dprShortFixed(raw.m1_dac1);
      m1.dac2   = dprShortFixed(raw.m1_dac2);
      m2.base   = dprShortFixed(raw.m2_base);
      m2.dac0   = dprShortFixed(raw.m2_dac0);
      m2.dac1   = dprShortFixed(raw.m2_dac1);
      m2.dac2   = dprShortFixed(raw.m2_dac2);

      /* print minfo to screen */
      printf("        ts %ld \n", m1.timestamp_plc0 );
      printf("    period %ld \n", m1.period_plc0 );
      printf("   m1.base %ld %.03fv \n", m1.base, map(m1.base, -32768, 32767, -10.0, 10.0));
      printf("   m1.dac0 %ld %.03fv \n", m1.dac0, map(m1.dac0, -32768, 32767, -10.0, 10.0));
      printf("   m1.dac1 %ld %.03fv \n", m1.dac1, map(m1.dac1, -32768, 32767, -10.0, 10.0));
      printf("   m1.dac2 %ld %.03fv \n", m1.dac2, map(m1.dac2, -32768, 32767, -10.0, 10.0));
      printf("   m2.base %ld %.03fv \n", m1.base, map(m2.base, -32768, 32767, -10.0, 10.0));
      printf("   m2.dac0 %ld %.03fv \n", m2.dac0, map(m2.dac0, -32768, 32767, -10.0, 10.0));
      printf("   m2.dac1 %ld %.03fv \n", m2.dac1, map(m2.dac1, -32768, 32767, -10.0, 10.0));
      printf("   m2.dac2 %ld %.03fv \n", m2.dac2, map(m2.dac2, -32768, 32767, -10.0, 10.0));

      usleep(500000);
   }
   sc_gotoxy( 1, 22 );

   return T3_ERR_NONE;
}

/*----------------------------------------------------------------------
**  do_show_vb() -  read and display DPR VB motor 1 data.
**----------------------------------------------------------------------
*/
int do_show_vb( char * par )
{
   int rc, i, loops;
   int addr;
   char * str_ptr;
   struct pmac_vb_cntl_t vbc;
   struct pm_VB_read_raw_t vbrd;
   struct pm_minfo_t mi;

   /* initialize motor info with scaling data */
   pm_mi_init_defaults( &mi );

   /* get star index */
   rc = parseInt_r( &loops, par, " ", &str_ptr  );
   if( rc != T3_ERR_NONE)
      loops = 1;

   for( i=0; i<loops; i++ )
   {
      sc_home();
      if( (i%10) == 0 )
         sc_clrscr(2);
      printf("loop %04d/%04d \n", i, loops ); 

      /* clear bits (actually bit 0) to refresh data */
      vbc.read_cntl = 0;
      pmac_mem_putblk( Fd, PMAC_VB_CNTL_ADDR, &vbc.read_cntl, sizeof(vbc.read_cntl));

      /* get vb control memory block */
      rc = pmac_mem_getblk( Fd, &vbc, PMAC_VB_CNTL_ADDR, sizeof(vbc));
      show_mem( &vbc, sizeof(vbc), "VB Cntl");
      printf("\n");
      printf("       read_cntl 0x%04x \n", vbc.read_cntl);
      printf("       timestamp %d \n", vbc.timestamp);
      printf("   read_buf_size %d \n", vbc.read_buf_size);
      printf(" read_buf_offset %d \n", vbc.read_buf_offset);
      printf("  write_buf_size %d \n", vbc.write_buf_size);
      printf("write_buf_offset %d \n", vbc.write_buf_offset);

      /* get vb read data */
      addr = PMAC_VB_DATA_ADDR + (PM_VB_READ_BUF_OFFSET*4) + (PM_VB_READ_BUF_NITEMS*4);
      rc = pmac_mem_getblk( Fd, &vbrd, addr, sizeof(vbrd));
      show_mem( &vbrd, sizeof(vbrd), "VB Read Data ");

      /* convert vb read raw data to minfo_t format */

      mi.timestamp_vb = vbc.timestamp;
      mi.read_cntl = vbc.read_cntl;
      mi.Ixx08     = dprShortFixed( vbrd.m1.Ixx08 );
      mi.Ixx09     = dprShortFixed( vbrd.m1.Ixx09 );
      mi.dpos     = dprLongFixed( vbrd.m1.dpos, 1.0/(mi.Ixx08*32.0)  );
      mi.pid_d    = dprShortFixed( vbrd.m1.pid_d );
      mi.pid_p    = dprShortFixed( vbrd.m1.pid_p );
      mi.pid_i    = dprShortFixed( vbrd.m1.pid_i );
      mi.adc1     = 0x0000ffff & (vbrd.m1.adc1 >> 8);
      mi.adc2     = 0x0000ffff & (vbrd.m1.adc2 >> 8);

      /* print minfo to screen */
      printf("      dpos %.3f \n", mi.dpos );
      printf("     pid_d %ld \n",  mi.pid_d);
      printf("     pid_p %ld \n",  mi.pid_p);
      printf("     pid_i %ld \n",  mi.pid_i);
      printf("      adc1 %ld %.2f  \n",  mi.adc1, map( mi.adc1, 0, 65536, -10.0,  10.0));
      printf("      adc2 %ld %.2f  \n",  mi.adc2, map( mi.adc2, 0, 65536, -10.0,  10.0));

      usleep(500000);
   }
   sc_gotoxy( 1, 22 );

   return T3_ERR_NONE;
}

/*----------------------------------------------------------------------
**  do_string_reverse() - calls pmac:ioctl(PMAC_CMD_STRING_REVERSE) 
**     Reverses a user sting. A driver test function.
**----------------------------------------------------------------------
*/
int do_string_reverse( char * par )
{
   int rc;
   char buf[80];
   char * before;

   before = "This is a string!";
   strxcpy( buf, before, sizeof(buf));

   rc = pmac_string_reverse( Fd, buf );
   printf(" rc = %d \n", rc );
   printf(" before = [%s] \n", before);
   printf(" after  = [%s] \n", buf);

   return 0;
}

/*----------------------------------------------------------------------
**  do_t1() - calls pmac:ioctl(PMAC_IOCTL_T1) 
**----------------------------------------------------------------------
*/
int do_t1( char * par )
{
   int rc;

   rc = pmac_t1( Fd );
   printf(" pmac_t1() returns %d\n", rc );
   return 0;
}

/*----------------------------------------------------------------------
**  do_t2() - calls pmac:ioctl(PMAC_IOCTL_T1) 
**----------------------------------------------------------------------
*/
int do_t2( char * par )
{
   int rc;

   rc = pmac_t2( Fd );
   printf(" pmac_t2() returns %d\n", rc );
   return 0;
}

/********************************************************************
**                 T3 APPLICATION specific COMMANDS                **
********************************************************************/

/*----------------------------------------------------------------------
**  do_pm_init() - Initialize the PMAC for TCS3 
**----------------------------------------------------------------------
*/
int do_pm_init( char * par )
{
   int rc;

   rc = pm_init( Fd );
   printf(" pm_init() returns %d\n", rc );
   return 0;
}

/*----------------------------------------------------------------------
**  do_pm_show() - Shows motor info data for motor 1 & 2.
**----------------------------------------------------------------------
*/
int do_pm_show( char * par )
{
   int rc, loops, i;
   struct pm_minfo_t m1;
   struct pm_minfo_t m2;
   char * str_ptr;

   pm_mi_init_defaults( &m1 );
   pm_mi_init_defaults( &m2 );

   /* get star index */
   rc = parseInt_r( &loops, par, " ", &str_ptr  );
   if( rc != T3_ERR_NONE)
      loops = 1;

   for( i=0; i<loops; i++ )
   {
#if 1
      sc_home();
      if( (i%10) == 0 )
         sc_clrscr(2);
#endif

      /* get Motor Information */
      rc = pm_mi_get_all( Fd, &m1, &m2 );
      if( rc )
      {
         sc_clrscr(2);
         printf("pm_mi_get() error. Is VB update enabled? \n");
         return T3_ERR_NONE;
      }
      else
      {
#if 0
         printf(" read_cntl 0x%x \n", m1.read_cntl );
         printf("vb_waitcnt %d \n", m1.vb_wait_cnt );
         printf("vb_waitsec %.6f \n", m1.vb_wait_sec );
#else 
         printf("PMAC MotorInformation %04d/%04d \n", i, loops ); 
         pm_show_print_motor_data(  &m1, "Motor1",  1, 3);
         pm_show_print_motor_data(  &m2, "Motor2", 40, 3);
#endif
      }
      usleep(200000);
   }

   return T3_ERR_NONE;
}

/*----------------------------------------------------------------------
**  pm_show_print_motor_data() - displays pm_minfo_t data to the screen       
**----------------------------------------------------------------------
*/
int pm_show_print_motor_data( 
   struct pm_minfo_t *m ,
   char * name, 
   int x, 
   int y
)
{
   long l;
   char bits[40];

   sc_gotoxy( x, y++);  printf(" ___________%s______________ ", name ); 

   /* motor data (from VB) */
   sc_gotoxy( x, y++); printf("     tm_vb %ld \n", m->timestamp_vb );
   sc_gotoxy( x, y++); printf(" read_cntl 0x%x \n", m->read_cntl );
   sc_gotoxy( x, y++); printf("vb_waitcnt %d \n", m->vb_wait_cnt );
   sc_gotoxy( x, y++); printf("vb_waitsec %.6f \n", m->vb_wait_sec );
   sc_gotoxy( x, y++); printf("     Ixx08 %.1f \n", m->Ixx08 );
   sc_gotoxy( x, y++); printf("     Ixx09 %.1f \n", m->Ixx09 );
   sc_gotoxy( x, y++); printf("      dpos %.3f \n", m->dpos );
   sc_gotoxy( x, y++); printf("     pid_d %ld \n",  m->pid_d);
   sc_gotoxy( x, y++); printf("     pid_p %ld \n",  m->pid_p);
   sc_gotoxy( x, y++); printf("     pid_i %ld \n",  m->pid_i);
   sc_gotoxy( x, y++); printf("      adc1 %ld %.3fv  \n",  m->adc1, 
      map( m->adc1, 0, 65535.0, -10.0,  10.0));
   sc_gotoxy( x, y++); printf("      adc2 %ld %.3fv  \n",  m->adc2, 
      map( m->adc2, 0, 65535.0, -10.0,  10.0));

   /* motor data (from MRB) */
   sc_gotoxy( x, y++); printf("    ts_mrb %ld ", m->timestamp_mrb);
   sc_gotoxy( x, y++); printf("      pdac %ld %.03fv ", m->pdac, 
      map( m->pdac, -32768, 32767, -10.0, 10.0));
   sc_gotoxy( x, y++); printf("      mpos %.3f ",  m->mpos);
   sc_gotoxy( x, y++); printf("      apos %.3f ",  m->apos);
   sc_gotoxy( x, y++); printf("      bpos %.3f ",  m->bpos);
   sc_gotoxy( x, y++); printf("     pferr %.3f ",  m->pferr);
   sc_gotoxy( x, y++); printf("      fvel %.3f ",  m->fvel );

   memcpy( &l, &m->sstatus, sizeof(l));
   sc_gotoxy( x, y++); printf("   SStatus 0x%08lx ", l );
   sc_gotoxy( x, y++); printf("       MSW %s ", btoss( bits, l >> 16));
   sc_gotoxy( x, y++); printf("       LSW %s ", btoss( bits, l));
   sc_gotoxy( x, y++); printf(" RSS=%d DSM=%d  SC=%d  FC=%d ", m->sstatus.rapid_spd_sel,
      m->sstatus.dac_sign_mag, m->sstatus.sw_capture, m->sstatus.fe_capture);
   sc_gotoxy( x, y++); printf("  HE=%d HWM=%d  PM=%d YEP=%d ", m->sstatus.handwheel_ena,
      m->sstatus.hw_mode, m->sstatus.phased_motor, m->sstatus.yenc_phase);
   sc_gotoxy( x, y++); printf("  US=%d  UP=%d  HS=%d  BR=%d ", m->sstatus.user_servo,
      m->sstatus.user_phase, m->sstatus.home_search, m->sstatus.block_request); 
   sc_gotoxy( x, y++); printf("  LS=%d DV0=%d DBE=%d  DW=%d ", m->sstatus.limit_stop,
      m->sstatus.desired_vel_0, m->sstatus.data_block_err, m->sstatus.dwelling);
   sc_gotoxy( x, y++); printf("  IE=%d  RP=%d  OL=%d AMP=%d ", m->sstatus.integrator_ena,
      m->sstatus.run_program, m->sstatus.open_loop, m->sstatus.amp_enabled);
   sc_gotoxy( x, y++); printf(" ALE=%d  PL=%d  NL=%d  ACT=%d ", m->sstatus.algo_ena,
      m->sstatus.pos_limit, m->sstatus.neg_limit, m->sstatus.activated);

   memcpy( &l, &m->mstatus, sizeof(l));
   sc_gotoxy( x, y++); printf("   MStatus 0x%08lx ", l );
   sc_gotoxy( x, y++); printf("       MSW %s ", btoss( bits, l >> 16));
   sc_gotoxy( x, y++); printf("       LSW %s ", btoss( bits, l));

   sc_gotoxy( x, y++); printf("  IP=%d WFE=%d FFE=%d  AF=%d ", m->mstatus.in_position, 
      m->mstatus.warn_ferr, m->mstatus.fatal_ferr, m->mstatus.amp_fault);
   sc_gotoxy( x, y++); printf("  BD=%d AIE=%d  IF=%d THF=%d ", m->mstatus.backlash_dir,
      m->mstatus.amp_i2t_err, m->mstatus.integral_ferr, m->mstatus.triger_home_flg);
   sc_gotoxy( x, y++); printf(" PFE=%d  HC=%d SOL=%d CSA=%d ", m->mstatus.phase_find_err,
      m->mstatus.home_complete, m->mstatus.stopped_on_limit, m->mstatus.cs_assigned);
   sc_gotoxy( x, y++); printf(" CSD=%d  CA=%d \n", m->mstatus.cs_def, m->mstatus.coord_sys);

   /* plc0 */
   sc_gotoxy( x, y++); printf("    ts.plc0 %ld \n", m->timestamp_plc0 );
   sc_gotoxy( x, y++); printf("period.plc0 %ld \n", m->period_plc0 );
   sc_gotoxy( x, y++); printf("       base %ld %03fv \n", m->base,
         map(m->base, -32768, 32767, -10.0, 10.0));
   sc_gotoxy( x, y++); printf("  dac 0/1/2 %ld,%ld,%ld \n", m->dac0, m->dac1, m->dac2);
   sc_gotoxy( x, y++); printf(" volt 0/1/2 %.03fv,%.03fv,%.03fv \n", 
         map(m->dac0, -32768, 32767, -10.0, 10.0),
         map(m->dac1, -32768, 32767, -10.0, 10.0),
         map(m->dac2, -32768, 32767, -10.0, 10.0));

   return T3_ERR_NONE;
}

/********************************************************************
**                           MISC COMMANDS                         **
********************************************************************/

/*----------------------------------------------------------------------
**  do_help() - set Quit flag to TRUE.
**----------------------------------------------------------------------
*/
int do_help( char * par )
{
   int i;

   printf("\n"
          "This program communicate with the PMAC driver. \n"
          "Summary of commands:\n");

   for( i=0; i<num_parse_item; i++)
   {
      printf("   %s\n", parse_table[i].desc);
   }
   printf("\n");
   return 0;
}


/*----------------------------------------------------------------------
**  do_quit() - set Quit flag to TRUE.
**----------------------------------------------------------------------
*/
int do_quit( char * par )
{
   Quit = TRUE;
   return 0;
}

/*----------------------------------------------------------------------
**  do_test() - test command
**----------------------------------------------------------------------
*/
int do_test( char * par )
{
   return 0;
}

/********************************************************************
**                     SUPPORTING FUNCTIONS                        **
********************************************************************/


/*----------------------------------------------------------------------
**  get_ulong() - prompt the user and reads a ulong from the keyboard.
**----------------------------------------------------------------------
*/
ulong get_ulong (  char *prompt, ulong cur_val, ulong min, ulong max )
{
   char buf[80];

   while( 1 )
   {
      printf( "%s (0x%lx): ", prompt, cur_val );
      gets( buf );
      unpad( buf, ' ' );
      if( strlen( buf) >= 1 )
         sscanf( buf, "%lx", &cur_val);

      if( INRANGE(min, cur_val, max))
      {
         return cur_val;
      }
      else
         printf("\007INVALID Input. Range is %lu to %lu.\n", min, max);
   }
}

/*----------------------------------------------------------------------
**  get_long() - prompt the user and reads a long from the keyboard.
**----------------------------------------------------------------------
*/
long get_long (  char *prompt, long cur_val, long min, long max )
{
   char buf[80];

   while( 1 )
   {
      printf( "%s (%ld): ", prompt, cur_val );
      gets( buf );
      unpad( buf, ' ' );
      if( strlen(buf) >= 1 )
         cur_val = atol(buf);

      if( INRANGE(min, cur_val, max))
      {
         return cur_val;
      }
      else
         printf("\007INVALID Input. Range is %ld to %ld.\n", min, max);
   }
}

/*----------------------------------------------------------------------
**  get_dbl() - prompt the user and reads a double from the keyboard.
**----------------------------------------------------------------------
*/
double get_dbl( char *prompt, double cur_val, double min, double max )
{
   char buf[80];

   while( 1 )
   {
      printf( "%s (%f): ", prompt, cur_val );
      gets( buf );
      unpad( buf, ' ' );
      if( strlen( buf) >= 1 )
         sscanf( buf, "%lf", &cur_val);
      if( INRANGE( min, cur_val, max))
         return cur_val;
      else
         printf("\007INVALID Input. Range is %g to %g.\n", min, max);
   }
}


/*----------------------------------------------------------------------
**  tm_start(), tm_end() - functions to time operations.
**----------------------------------------------------------------------
*/
static struct timespec start;

void tm_start( void )
{
   clock_gettime( CLOCK_REALTIME, &start );
}

double tm_end( void )
{
   struct timespec end;
   double d;
   long nsec;

   clock_gettime( CLOCK_REALTIME, &end );

   d  = (end.tv_sec - start.tv_sec);
   nsec  = end.tv_nsec - start.tv_nsec;
   d += nsec/1000000000.0; // convert nsec to sec and add in (OK if negative)

   return d;
}


/*--------------------------------------------------------------
**  show_mem() - print a dump of RAM to stdout.
**--------------------------------------------------------------
*/
void show_mem( void * mem, int nbytes, char * desc )
{
   int i;
   unsigned char *c;

   c = (char*) mem;

   // heading
   printf("\n       ");
   for( i=0; i<16; i++ )
      printf("%02x ", i);
   printf("\n-- %s -nbytes=%d ---------------------", desc, nbytes);

   for( i=0; i<nbytes; i++ )
   {
      if( (i%16)==0 ) printf("\n0x%04x ", i);
      printf("%02x ", c[i]);
   }
   printf("\n");
}

/*--------------------------------
** Some screen controls functions
**
**   sc_clrscr( p )   - clears the screen.
**   sc_clrln( p )    - clears the current line.
**      0 - cursor to end.
**      1 - beg to cursor.
**      2 - entire line or screen.
**   sc_gotoxy( x, y) - Moves the cursor. 1,1 is upper-left.
**   sc_home()        - Move the curser to 1,1
**--------------------------------
**/
int sc_clrscr( int p )
{
   printf("\033[%dJ", p);
   return T3_ERR_NONE;
}
int sc_clrln( int p )
{
   printf("\033[%dK", p);
   return T3_ERR_NONE;
}

int sc_gotoxy( int x, int y)
{
   printf("\033[%d;%dH", y, x);
   return T3_ERR_NONE;
}
int sc_home(void )
{
   printf("\033[H");
   return T3_ERR_NONE;
}



/*---------------------------------------------------------------------------------
** stio_cmd() - SimTac IO - does a command/reply from simtac
**---------------------------------------------------------------------------------
*/
int stio_cmd ( char * cmd, char * reply, int reply_size )
{
   int rc;
   char buf[200];
   FILE * fp = NULL;

   sprintf( buf, "/home/tcs3/src/tcs3/simtac/stio '%s' ", cmd);
   if( (fp = popen( buf, "r")) == NULL )
   {
     return T3_ERR_NOT_AVAILABLE;
   }
   rc = T3_ERR_NONE;

   /*-----------------------------------
   ** tcs should return -OK
   */
   buf[0] = 0;
   if( fgets(buf, sizeof(buf), fp ) != NULL )
   {
      strxcpy( reply, buf, reply_size );  // copy reply to caller
   }

ldone:
   pclose(fp);
   return rc;
}


